
#
#  We implement the algorithm described in the last section of the paper.
#

#  We focus on the 2d case, and first define arrays containing the random conductances. We will use two arrays, one for the horizontal edges ("horiz"), and one for the vertical ones ("verti"). Each will be defined by calling this function.

conduct(N::Int) = map( (x) -> x < 0.5 ? 1. : 9., rand(N,N))

# We see that we chose the environment to be 1 or 9 with prob. 1/2 each. 

# Our conventions for sites and edges are as follows. Sites are indexed by [2,N]^2. Functions defined on sites are defined on [1,N+1]^2, and completed to 0 for convenience; functions on edges are defined on [1,N]^2. A call to horiz(i,j) returns the conductance of the edge between the sites (i,j) and (i,j+1). A call to verti(i,j) returns the conductance of the edge between (i,j) and (i+1,j). The Neumann boundary condition could be implemented by imposing that horiz[:,1] = 0, horiz[:,N] = 0, verti[1,:] = verti[N,:] = 0. The Dirichlet boundary condition is imposed by checking that the values at [1,:], [N+1,:], [:,1] and [:,N+1] are zero. We implement the latter, since it is marginally simpler.


# We next define the matrix pii, i.e. the invariant measure. We just need to be careful about the boundary condition.

function make_pii(horiz::Array{Float64,2},verti::Array{Float64,2}) 
	N = size(horiz,1)
	pii = Array{Float64}(N+1,N+1)
	pii[1,:] = 0.
	pii[N+1,:] = 0.
	pii[:,1] = 0.
	pii[:,N+1] = 0.
	for col = 2:N , row = 2:N
		pii[row,col] = horiz[row,col] + horiz[row,col-1] + verti[row,col] + verti[row-1,col]
	end
	return pii
end

# We store the value of the local drift into an array, fixing first the vector \xi.

xi1 = 1./sqrt(2.) ; xi2 = 1./sqrt(2.) ; 

function make_drift(horiz::Array{Float64,2}, verti::Array{Float64,2}, pii::Array{Float64,2}) 
	N = size(horiz,1)
	drift = Array{Float64}(N+1,N+1)
	drift[1,:] = 0.
	drift[N+1,:] = 0.
	drift[:,1] = 0.
	drift[:,N+1] = 0.
	for col = 2:N , row = 2:N
		drift[row,col] = 1./pii[row,col] * ( xi1 * (horiz[row,col] - horiz[row,col-1]) + xi2 * (verti[row,col] - verti[row-1,col]) )
	end
	return drift
end

# We now let the heat flow on data for one time step. 

function heat_flow(horiz::Array{Float64,2}, verti::Array{Float64,2}, pii::Array{Float64,2}, data::Array{Float64,2},N)
#	N = size(horiz,1)
	flow = copy(data)
	for col = 2:N , row = 2:N
		flow[row,col] = 
		0.5 * data[row,col] +  0.5/pii[row,col] * (horiz[row,col-1] * data[row,col-1] + horiz[row,col] * data[row,col+1] + verti[row,col] * data[row+1,col] + verti[row-1,col] * data[row-1,col])
	end
	return flow
end

#  We are now ready to run the algorithm. The quantity D_L in the pdf file is approx(L).



function approx(L::Int)
	approxim = 0.
	running_sum = 0.
	blayer = 5 * (max(L,1)) * floor(Int,sqrt(1.+ 2.^(float(L)/2.)))  # the size of the boundary layer
	N = 2^L + 2*blayer # the size of the box, with the boundary layer. 
	left_bdy = blayer + 1
	right_bdy = blayer + 2^L
	horiz = conduct(N)
	verti = conduct(N)
	pii = make_pii(horiz,verti)
	data = make_drift(horiz,verti,pii)
	data1 = heat_flow(horiz,verti,pii,data,N)
	for l = 0:(L-1)
		for k = (2^l-1):(2^(l+1)-2)
			running_sum += mean(pii[left_bdy:right_bdy,left_bdy:right_bdy] .*(data[left_bdy:right_bdy,left_bdy:right_bdy].^2. + data[left_bdy:right_bdy,left_bdy:right_bdy] .* data1[left_bdy:right_bdy,left_bdy:right_bdy]))
			data = copy(data1)
			data1 = heat_flow(horiz,verti,pii,data,N)		
		end
		approxim += running_sum/2.
		running_sum = 0.
				# update the size of the matrices 
		neweffectivesize = ceil(Int,2.^(float(L-((l+1)/2)))) 
		newN = neweffectivesize + 2*blayer
		newleft = max(1  , floor(Int,(N-newN)/2) )
		newright = newleft + newN - 1
		horiz = horiz[newleft:newright, newleft:newright]
		verti = verti[newleft:newright, newleft:newright] 	# keep in mind the funny fact that "sites" arrays are a bit longer
		pii = pii[newleft:(newright+1), newleft:(newright+1)]
		data = data[newleft:(newright+1), newleft:(newright+1)]
		data1 = data1[newleft:(newright+1), newleft:(newright+1)]
		right_bdy =  neweffectivesize + blayer
		N = newN
	end
	return approxim
end
