# This document is split into several parts:

# --  base contains
#	conduct
#	make_horiz
#	make_verti
#	divagrad
#	divaxi
#	
# --  cholesky contains
# 	cholesky
# 	icholesky
# 	MyPreconditionerType
# 	new methods for A_ldiv_B!
# 	some garbage
# 
# --  classical contains
#	base_classical
# 	classical
# 	base_classical_pre
# 	classical_pre (the preconditioned version)
# 	
# 
# --  itresolv contains
# 	itresolv	


# The "classical" part requires functions in "base" and "cholesky"; the "itresolv" part requires functions in "base". The "itresolv" function implements the method introduced in the paper. 



# The code was tested on Julia 0.6 and requires the following package:

import Base.LinAlg.A_ldiv_B!

using IterativeSolvers





###################
#
#
#	Base 
#
#
###################



# We store here functions used for the computation of homogenized coefficients based on "resolvent methods", namely the classical technique of resolving Poisson equations on a single scale (cf. classical.jl) and the iterated resolvents technique (cf. itresolv.jl). 

# Our conventions for the enumeration of sites and edges are as follows. Sites are indexed by [1,N]^2. Functions defined on the edges are represented by a pair of arrays indexed by [1,N]x[1,N-1] and [1,N-1]x[1,N] respectively. In particular, we will use two arrays, horiz and verti, to represent the value of the conductivities. A call to horiz[i,j] will return the conductance of the edge between the sites (i,j) and (i,j+1). A call to verti[i,j] will return the conductance of the edge between (i,j) and (i+1,j). 

#  We define the functions allowing to create these arrays.

conduct(N::Int) = map( (x) -> x < 0.5 ? 1. : 9. , rand(N,N) )

make_horiz(N::Int) = conduct(N)[:,1:end-1]

make_verti(N::Int) = conduct(N)[1:end-1,:]

# We see that we chose the environment to be 1 or 9 with prob. 1/2 each. 

# In order to define the matrix corresponding to - \nabla \cdot \a(x) \nabla, we need to use a linear ordering of the sites of the graph. Elements of [1,N]^2 are enumerated according to the mapping (i,j) -> i + N*(j - 1). We implement a Neumann-type boundary condition. We will add a massive term in all of our systems later, which will ensure solvability. We choose the sign convention so that the operator is positive definite.

function divagrad(horiz::Array{Float64,2},verti::Array{Float64,2}) 
	N = size(horiz,1)
	#
	# we first build the non-diagonal part; the horizontal edges are easier to deal with since we always find a neighbor when we add N to a site index.
	A = -sparse(1:(N^2 - N), (N+1):N^2, horiz[:], N^2, N^2) 
	#
	# For the vertical ones, we pay attention that adding 1 to a vertex index does not necessarily give us a neighbor; so we add a last line of zeros before use.
	trim_verti = vcat(verti,zeros(N)')
	A = A - sparse( 1:(N^2-1), 2:N^2, trim_verti[:][1:end-1], N^2, N^2)
	#
	# We then add the lower part of this so-far triangular matrix and the diagonal
	A = A + A' + sparse(1:(N^2 - N), 1:(N^2 - N), horiz[:], N^2, N^2) + sparse(1:(N^2 - 1), 1:(N^2 - 1), trim_verti[:][1:end-1], N^2, N^2) + sparse(1:N^2, 1:N^2, hcat(zeros(N),horiz)[:] + vcat(zeros(N)',verti)[:] , N^2, N^2)
	#
	return A
end

# We now write a function that computes the right side of the system we want to solve, that is, \nabla \cdot \a(x) \xi. For simplicity, we only consider the case when \xi is the unit horizontal vector here. 

function divaxi(horiz::Array{Float64,2},verti::Array{Float64,2}) 
	N = size(horiz,1)
	#
	return (hcat(horiz,zeros(N)) - hcat(zeros(N),horiz))[:]
end










###################
#
#
#	Cholesky
#
#
###################



# We define Cholesky and incomplete Cholesky factorization algorithms, as well as a method to pass incomplete Cholesky factorizations as preconditioners to the conjugate gradient algorithm.

# We first define a standard "jki" Cholesky factorization algorithm, and then proceed to "sparsify" it. This Cholesky algorithm for dense matrices is not used, it only serves to illustrate the structure and "justify" the sparsified version below.

function cholesky(A::Array{Float64,2})

N = size(A,1)

L = copy(A)

for j = 1:N
	L[j,j] = sqrt(L[j,j])
	for i = (j+1):N
		for k = 1:(j-1)
			L[i,j] = L[i,j] - L[i,k]*L[j,k]
		end
	end
	for i = (j+1):N
		L[i,j] = L[i,j]/L[j,j]
		L[i,i] = L[i,i] - L[i,j]^2
	end
end

for i = 1:N, j = (i+1):N
	L[i,j] = 0
end

return L

end


# We now define an incomplete Cholesky factorization. !Important!: this algorithm is *not* a generic incomplete Cholesky method; we use the specific structure of the matrices we are interested in to speed up the computation. The algorithm follows that for dense matrices above, except that it only iterates over non-zero entries of the original matrix A. (In particular, every null entry of A is a null entry of L.)



function icholesky(A::SparseMatrixCSC{Float64,Int64})

N = size(A,1) 

L = copy(A)

(I,J,V) = findnz(L)
nz = length(I)

# We check which indices correspond to a change of column: the elements of L belonging to column j correspond to "linear" indices ranging between indexcol[j] and indexcol[j+1]-1.

indexcol = zeros(Int,N+1)

indexcol[1] = 1
indexcol[N+1] = nz+1

for col = 2:N
	indexcol[col] = findnext(x -> x == col, J, indexcol[col-1])
end

for j = 1:N
	L[j,j] = sqrt(L[j,j])
	for l = indexcol[j]:(indexcol[j+1]-1)
		i = I[l]
		j = J[l]
		if i > j
			# we compute (L[i,1:(j-1)] * L[j,1:(j-1)]')[1,1] using that the only non-zero element in this scalar product is at the (j-1)-th position (indeed, the only non-zero elements of L[i,1:(j-1)] have indices i-sqrt(N) and i-1, and similarly for L[j,1:(j-1)], and moreover j < i, so the scalar product is nonzero only if the indices i-sqrt(N) and j-1 are equal). 
			j > 1 ? scalprod = L[i,j-1] * L[j,j-1] : scalprod = 0.
			L[i,j] = L[i,j] - scalprod # - (L[i,1:(j-1)] * L[j,1:(j-1)]')[1,1] 
		end
	end
	for l = indexcol[j]:(indexcol[j+1]-1)
		i = I[l]
		j = J[l]
		if i > j
			L[i,j] = L[i,j]/L[j,j]
			L[i,i] = L[i,i] - L[i,j]^2.
		end
	end
	
end

return LowerTriangular(L)

end


# We now define a suitable method to pass an incomplete Cholesky factorization as a preconditioner to the conjugate gradient algorithm.

struct MyPreconditionerType
           lower::LowerTriangular
       end

MyPreconditionerType(L::SparseMatrixCSC) = MyPreconditionerType(LowerTriangular(L))


# Performs x <- L' \ (L \ x) in-place with backward & forward substitutions
# A_ldiv_B!(P::MyPreconditionerType, x) = Ac_ldiv_B!(P.lower, A_ldiv_B!(P.lower, x))
# This has a problem because Ac_ldiv_B! is not encoded properly
A_ldiv_B!(P::MyPreconditionerType, x) = A_ldiv_B!(UpperTriangular(P.lower'), A_ldiv_B!(P.lower, x))
A_ldiv_B!(y, P::MyPreconditionerType, x) = A_ldiv_B!(P, copy!(y, x))


# For L = icholesky(A), we can then solve for x in A*x = b by calling

# x, history = cg(A, b, Pl = MyPreconditionerType(L))










###################
#
#
#	Classical
#
#
###################







# the following function computes the homogenized coefficients using the classical approach of solving for the corrector (in the horizontal direction) with a massive term mu (chosen by the user) on a box of side length N + a boundary layer of side length 5 \mu^{-1/2} \log_2 (\mu^{-1/2}), and then returns the averaged energy inside the box (outside of the boundary layer).


function base_classical(N::Int, mu::Float64)

if N <= 1 
	N = 2 
end

# we first define the size of the boundary layer

blayer = floor(Int,5. * max(1.,mu^(-0.5)) * max(1.,log(mu^(-0.5))/log(2.))) 
left_bdy = blayer + 1 
right_bdy = blayer + N
L = right_bdy + blayer

# we then define the conductances, the relevant arrays of the linear system, and solve it.

horiz = make_horiz(L) 
verti = make_verti(L)
A = divagrad(horiz,verti)

phi = reshape(IterativeSolvers.cg(mu * speye(L^2) + A,divaxi(horiz,verti)),L,L) 

# we now define the gradient of the corrector phi and return its average energy.

dhorizphi = phi[:,2:end] - phi[:,1:end-1] 
dvertiphi = phi[2:end,:] - phi[1:end-1,:] 

energy_h = (ones(L,L-1) + dhorizphi) .* horiz .* (ones(L,L-1) + dhorizphi) 
energy_v = dvertiphi .* verti .* dvertiphi 

return mean(energy_h[left_bdy:right_bdy,left_bdy:right_bdy-1]) + mean(energy_v[left_bdy:right_bdy-1,left_bdy:right_bdy])

end


# and now, given an integer n, we solve 2^(nd/2) problems each defined on a box of side length 2^(n/2) and with massive term mu = 2^(-n) (and here d = 2).

function classical(n::Int)
thesum = 0.
for i = 1:2^n
	thesum += base_classical(ceil(Int,2.^(n/2.)), 2.^(-n*1.))
end
return thesum/(2.^(n*1.))
end


# And now the same procedure with preconditioning

function base_classical_pre(N::Int, mu::Float64)

if N <= 1 
	N = 2 
end

# we first define the size of the boundary layer

blayer = floor(Int,5. * max(1.,mu^(-0.5)) * max(1.,log(mu^(-0.5))/log(2.))) 
left_bdy = blayer + 1 
right_bdy = blayer + N
L = right_bdy + blayer

# we then define the conductances, the relevant arrays of the linear system, and solve it.

horiz = make_horiz(L) 
verti = make_verti(L)
A = divagrad(horiz,verti)
Amu = mu * speye(L^2) + A
P = MyPreconditionerType(icholesky(Amu)) 

phi = reshape(IterativeSolvers.cg(Amu,divaxi(horiz,verti), Pl=P),L,L) 

# we now define the gradient of the corrector phi and return its average energy.

dhorizphi = phi[:,2:end] - phi[:,1:end-1] 
dvertiphi = phi[2:end,:] - phi[1:end-1,:] 

energy_h = (ones(L,L-1) + dhorizphi) .* horiz .* (ones(L,L-1) + dhorizphi) 
energy_v = dvertiphi .* verti .* dvertiphi 

return mean(energy_h[left_bdy:right_bdy,left_bdy:right_bdy-1]) + mean(energy_v[left_bdy:right_bdy-1,left_bdy:right_bdy])

end




function classical_pre(n::Int)
thesum = 0.
for i = 1:2^n
	thesum += base_classical_pre(ceil(Int,2.^(n/2.)), 2.^(-n*1.))
end
return thesum/(2.^(n*1.))
end














###################
#
#
#	Iterated resolvents
#
#
###################


# We implement the method exposed in the paper. 

function itresolv(n::Int)

sum_over_k = 0.

mu = 1.
N = 2^n
blayer = floor(Int,5. * max(1.,mu^(-0.5)) * max(1.,log(mu^(-0.5))/log(2.))) 
left_bdy = blayer + 1 
right_bdy = blayer + N
L = right_bdy + blayer
horiz = make_horiz(L) 
verti = make_verti(L)
A = divagrad(horiz,verti)
b2 = reshape(divaxi(horiz,verti),L,L)

for k = 0:(n-1)

	# we shift the last iteration to b1 and compute the new resolvent in b2
	b1 = b2 + 0.
	b2 = reshape(IterativeSolvers.cg(mu * speye(L^2) + A, mu * b1[:], maxiter = size(A,2)),L,L)
	
	b1r = b1[left_bdy:right_bdy,left_bdy:right_bdy]
	b2r = b2[left_bdy:right_bdy,left_bdy:right_bdy]
	
	sum_over_k += 1./mu * mean(b1r.*b2r + b2r .^2.)
		
	# and now we reshape our arrays
	
	mu = mu/2.
	N = floor(Int, 2.^(n-0.5*(k+1.)))
	blayer = min(floor(Int,5. * max(1.,mu^(-0.5)) * max(1.,log(mu^(-0.5))/log(2.))), floor(Int, (L-N)/2.)) 
	left_bdy = blayer + 1 
	right_bdy = blayer + N
	oldL = L + 0
	L = min(right_bdy + blayer, oldL)
	gapL = floor(Int, (oldL-L)/2.)
	
	b1 = b1[gapL+1:gapL+L,gapL+1:gapL+L]
	b2 = b2[gapL+1:gapL+L,gapL+1:gapL+L]
	horiz = horiz[gapL+1:gapL+L,gapL+1:gapL+L-1]
	verti = verti[gapL+1:gapL+L-1,gapL+1:gapL+L]
	A = divagrad(horiz,verti)
	
end

return sum_over_k

end
